unit hiZoomer;

interface

uses
  Kol,Windows,Share,Debug,Win;

type
  TZoomerPressedType = (zptNone, zptDown, zptUp,zptPageDown, zptPageUp,
                           zptSlider, zptSliderDown, zptSliderUp);
  TZoomerNotify = procedure(Sender: PObj; Position, Range: Integer)of object;

  THIZoomer = class(THIWin)
  private
    FGlyphs:             KOL.PBitmap;
    FBmp:                KOL.PBitmap;
    FTimer:              PTimer;

    FModelPosition:      Integer;
    FModelRange:         Integer;
    FModelMinRange:      Integer;
    FModelMaxRange:      Integer;
    FShowSlider:         Boolean;

    FOnChanged:          TZoomerNotify;

    FTimeStartRepeat:    Cardinal;
    FRepeatDelay:        Cardinal;
    FSliderX:            Integer;
    FSliderWidth:        Integer;
    FPressedType:        TZoomerPressedType;
    FPressedX:           Integer;
    FPressedSliderX:     Integer;
    FPressedSliderWidth: Integer;
    FPressedPosition:    Integer;
    FPressedRange:       Integer;

    procedure SetModelPosition(Value: Integer);
    procedure SetModelRange(Value: Integer);
    procedure SetModelMinRange(Value: Integer);
    procedure SetModelMaxRange(Value: Integer);

    function  ModelToSlider: Boolean;
    procedure VerifyData;
    procedure VerifySliderWidth;

    function  GetPressedType(X: Integer): TZoomerPressedType;
    function  GetLineSize: Integer;
    function  GetPageSize: Integer;

    procedure OnTimer(Sender: PObj);
    procedure DoChanged;

    procedure ChangePositionByPressedType;
    procedure CreateBrushPattern(Bitmap: KOL.PBitmap);
    procedure Paint(Sender: PControl; DC: HDC);
    procedure Resize(Obj: PObj);
    procedure onMouseDown(Sender: PControl; var Mouse: TMouseEventData);
    procedure onMouseUp(Sender: PControl; var Mouse: TMouseEventData);
    procedure onMouseMove(Sender: PControl; var Mouse: TMouseEventData);
  public
    _prop_Gradient: boolean;
    _event_OnRange: THI_Event;
    _event_onPosition: THI_Event;

    _prop_Position,_prop_Range,
    _prop_MinRange,_prop_MaxRange: integer;

    procedure Init;
    destructor  Destroy; override;

    procedure   SetData(aMinRange, aMaxRange, aPosition, aRange: Integer);
    procedure   HideSlider;

    property    Position: Integer read FModelPosition write SetModelPosition;
    property    Range: Integer read FModelRange write SetModelRange;
    property    MinRange: Integer read FModelMinRange write SetModelMinRange;
    property    MaxRange: Integer read FModelMaxRange write SetModelMaxRange;

    property OnChanged: TZoomerNotify read FOnChanged write FOnChanged;
  end;


implementation

const
  WS  = 6; //      
  HS  = 1; //         
  XSR = 2; //     
  YSR = 2; //     
  GCR = 1; //    
  WR  = 2; //  
  FSliderMinWidth = 2 * WS + 3 * WR + 4 * GCR + 2;

{ THIZoomer }

procedure THIZoomer.SetModelPosition(Value: Integer);
begin
  if (Value + FModelRange) > FModelMaxRange then
    Value := FModelMaxRange - FModelRange;
  if Value < 0 then
    Value := 0;
  if FModelPosition <> Value then begin
    FModelPosition := Value;
    if ModelToSlider then
      Control.Invalidate;
  end;
end;

procedure THIZoomer.SetModelRange(Value: Integer);
var
  p: Integer;
begin
  p := FModelPosition;
  if (p + Value) > FModelMaxRange then
    Value := FModelMaxRange - p;
  if Value > FModelMaxRange then
    Value := FModelMaxRange;
  if Value < FModelMinRange then
    Value := FModelMinRange;
  if (p + Value) > FModelMaxRange then
    p := FModelMaxRange - Value;
  if p < 0 then
    p := 0;
  if (FModelRange <> Value) or (FModelPosition <> p) then begin
    FModelRange    := Value;
    FModelPosition := p;
    if ModelToSlider then
      Control.Invalidate;
  end;
end;

procedure THIZoomer.SetModelMinRange(Value: Integer);
var
  p, r: Integer;
begin
  p := FModelPosition;
  r := FModelRange;
  if Value < 1 then
    Value := 1;
  if Value > (FModelMaxRange - 1) then
    Value := FModelMaxRange - 1;
  if r < Value then
    r := Value;
  if (p + r) > FModelMaxRange then
    p := FModelMaxRange - r;
  if p < 0 then
    p := 0;
  if (FModelMinRange <> Value) or
     (FModelPosition <> p) or
     (FModelRange <> r) then
  begin
    FModelMinRange := Value;
    FModelPosition := p;
    FModelRange    := r;
    if ModelToSlider then
      Control.Invalidate;
  end;
end;

procedure THIZoomer.SetModelMaxRange(Value: Integer);
var
  p, r: Integer;
begin
  p := FModelPosition;
  r := FModelRange;
  if Value < 2 then
    Value := 2;
  if Value < (FModelMinRange + 1) then
    Value := FModelMinRange + 1;
  if r > Value then
    r := Value;
  if (p + r) > Value then
    p := Value - r;
  if p < 0 then
    p := 0;
  if (FModelMaxRange <> Value) or
     (FModelPosition <> p) or
     (FModelRange <> r) then
  begin
    FModelMaxRange := Value;
    FModelPosition := p;
    FModelRange    := r;
    if ModelToSlider then
      Control.Invalidate;
  end;
end;

function THIZoomer.ModelToSlider: Boolean;
var
  x, w, mw, mp, mr: Integer;
begin
  mw := Control.Width - 2 * Control.Height;
  if mw > 0 then
    mr := MulDiv(FModelMaxRange, FSliderMinWidth, mw)
  else
    mr := FModelMinRange;
  if FModelRange > mr then
    mr := FModelRange;
  mp := FModelMaxRange - mr;
  x  := FSliderX;
  w  := FSliderWidth;
  FSliderWidth := MulDiv(mr, mw, FModelMaxRange);
  VerifySliderWidth;
  FSliderX := Control.Height + MulDiv(FModelPosition, mw - FSliderWidth, mp);
  if (FSliderX + FSliderWidth) > (Control.Width - Control.Height) then
    FSliderX := Control.Width - Control.Height - FSliderWidth;
  if FSliderX < Control.Height then
    FSliderX := Control.Height;
  Result := (x <> FSliderX) or (w <> FSliderWidth);
end;

procedure THIZoomer.VerifyData;
begin
  if FModelMinRange < 1 then
    FModelMinRange := 1;
  if FModelMaxRange < (FModelMinRange + 1) then
    FModelMaxRange := FModelMinRange + 1;
  if FModelRange > FModelMaxRange then
    FModelRange := FModelMaxRange;
  if FModelRange < FModelMinRange then
    FModelRange := FModelMinRange;
  if (FModelPosition + FModelRange) > FModelMaxRange then
    FModelPosition := FModelMaxRange - FModelRange;
  if FModelPosition < 0 then
    FModelPosition := 0;
end;

procedure THIZoomer.VerifySliderWidth;
var
  mw, mr: Integer;
begin
  mw := Control.Width - 2 * Control.Height;
  if FSliderWidth > mw then
    FSliderWidth := mw;
  if FSliderWidth < FSliderMinWidth then
    FSliderWidth := FSliderMinWidth;
  mr := MulDiv(FModelMinRange, mw, FModelMaxRange);
  if FSliderWidth < mr then
    FSliderWidth := mr;
end;

function THIZoomer.GetPressedType(X: Integer): TZoomerPressedType;
const
  G = 4;
begin
  if X < Control.Height then
    Result := zptDown
  else if X >= (Control.Width - Control.Height) then
    Result := zptUp
  else if X < FSliderX then
    Result := zptPageDown
  else if X >= (FSliderX + FSliderWidth) then
    Result := zptPageUp
  else if X < (FSliderX + G) then
    Result := zptSliderDown
  else if X >= (FSliderX + FSliderWidth - G) then
    Result := zptSliderUp
  else
    Result := zptSlider;
end;

function THIZoomer.GetLineSize: Integer;
begin
  Result := FModelRange div 10;
  if Result = 0 then
    Result := 1;
end;

function THIZoomer.GetPageSize: Integer;
begin
  Result := FModelRange;
end;

procedure THIZoomer.OnTimer(Sender: PObj);
begin
  if GetTickCount > FTimeStartRepeat then
    ChangePositionByPressedType;
end;

procedure THIZoomer.DoChanged;
begin
  //if Assigned(FOnChanged){ and FShowSlider} then
    //FOnChanged(@Self, FModelPosition, FModelRange);
  _hi_onEvent(_event_onPosition, FModelPosition);
  _hi_onEvent(_event_OnRange, FModelRange);
end;

procedure THIZoomer.ChangePositionByPressedType;
var
  p: Integer;
begin
  p := FModelPosition;
  case FPressedType of
    zptDown:     SetModelPosition(FModelPosition - GetLineSize);
    zptUp:       SetModelPosition(FModelPosition + GetLineSize);
    zptPageDown: SetModelPosition(FModelPosition - GetPageSize);
    zptPageUp:   SetModelPosition(FModelPosition + GetPageSize);
  end;
  if p <> FModelPosition then
    DoChanged;
end;

procedure THIZoomer.Paint;
var
  r:                TRect;
  w, h:             Integer;
  xbL, xbR, xL, xR: Integer;
  x, xC, YCR:       Integer;
  flags:            Cardinal;
  //c:                TColor;
  //y:                Integer;

  procedure DrawTopLine(aX0, aX1: Integer);
  begin
    FBmp.Canvas.Pen.Color := clBtnFace;
    FBmp.Canvas.MoveTo(0, 0);
    FBmp.Canvas.LineTo(w, 0);
  end;

  procedure GradientRect(gR: TRect);
  var
    c: TColor;
    y: Integer;
  begin

  //  
  //TZoomerPressedType = (zptNone, zptDown, zptUp,zptPageDown, zptPageUp,
  //                         zptSlider, zptSliderDown, zptSliderUp);
    if (FPressedType = zptDown) or (FPressedType = zptUp) or (FPressedType = zptSlider)then
    begin
      c := clWhite;
       for y := gR.Top + 1 to gR.Bottom - 1 do begin
         FBmp.Canvas.Pen.Color := c;
         FBmp.Canvas.MoveTo(gR.Left, y);
         FBmp.Canvas.LineTo(gR.Right, y);
         c := c - $030303;
         if c > $FFFFFF then
           c := $FFFFFF;
       end;
    end;

    if (FPressedType = zptNone) or (FPressedType = zptSliderDown) or
    (FPressedType = zptSliderUp) or (FPressedType = zptPageDown) or
    (FPressedType = zptPageUp) then
    begin
      c := $C0C0C0;
       for y := gR.Bottom - 1 downto gR.Top + 1 do begin
         FBmp.Canvas.Pen.Color := c;
         FBmp.Canvas.MoveTo(gR.Left, y);
         FBmp.Canvas.LineTo(gR.Right, y);
         c := c + $060606;
         if c > $FFFFFF then
           c := $FFFFFF;
       end;
    end;
  end;

begin
  w   := Control.Width;
  h   := Control.Height;
  xbL := h;
  xbR := w - h;
  xL  := FSliderX;
  xR  := xL + FSliderWidth;
  xC  := xL + ((xR - xL) div 2) - 1;
  YCR := (h - (3 * WR + 2 * GCR - 1)) div 2;

  if not FShowSlider then begin
    r := MakeRect(0, 0, w, h);
    FBmp.Canvas.Brush.BrushBitmap := FGlyphs.Handle;
    FBmp.Canvas.FillRect(r);
    FBmp.Canvas.Brush.BrushBitmap := 0;
    DrawTopLine(0, w);
    Exit;
  end;

  //   
  if w >= h then begin
    flags := DFCS_SCROLLLEFT;
    if FPressedType = zptDown then
    begin
      flags := flags or DFCS_PUSHED or DFCS_FLAT;
      r := MakeRect(0, 1, xbL, h);
    end else
      r := MakeRect(0, 0, xbL, h);
    if _prop_Gradient then
    begin
      GradientRect(r);
      FBmp.Canvas.Brush.BrushStyle := bsClear;
      FBmp.Canvas.DrawText('3',r,DT_CENTER or DT_TOP);
    end else
      DrawFrameControl(FBmp.Canvas.Handle, r, DFC_SCROLL, flags);
  end;

  //   
  if w >= (2 * h) then begin
    flags := DFCS_SCROLLRIGHT;
    if FPressedType = zptUp then
    begin
      flags := flags or DFCS_PUSHED or DFCS_FLAT;
      r := MakeRect(xbR, 1, w, h);
    end else
      r := MakeRect(xbR, 0, w, h);
    if _prop_Gradient then
    begin
      GradientRect(r);
      FBmp.Canvas.Brush.BrushStyle := bsClear;
      FBmp.Canvas.DrawText('4',r,DT_CENTER or DT_TOP);
    end else
      DrawFrameControl(FBmp.Canvas.Handle, r, DFC_SCROLL, flags);
  end;

  //   
  if xL > xbL then begin
    r := MakeRect(xbL, 0, xL, h);
    if FPressedType = zptPageDown then
      FBmp.Canvas.Brush.Color := clBtnShadow
    else
      FBmp.Canvas.Brush.BrushBitmap := FGlyphs.Handle;
    FBmp.Canvas.FillRect(r);
    DrawTopLine(xbL, xL);
  end;

  //   
  if xR < xbR then begin
    r := MakeRect(xbR, 0, xR, h);
    if FPressedType = zptPageUp then
      FBmp.Canvas.Brush.Color := clBtnShadow
    else if FBmp.Canvas.Brush.BrushBitmap = 0 then
      FBmp.Canvas.Brush.BrushBitmap := FGlyphs.Handle;
    FBmp.Canvas.FillRect(r);
    DrawTopLine(xbR, xR);
  end;
  FBmp.Canvas.Brush.BrushBitmap := 0;

  if (w - 2 * h) > FSliderMinWidth then begin
    //   
    r := MakeRect(xL, 0, xR, h);
    FBmp.Canvas.Brush.Color := COlor2RGB(Control.Color);//clRed;
    FBmp.Canvas.FillRect(r);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT or BF_MIDDLE);
    //   
    r := MakeRect(xL + WS, HS, xR - WS, h - HS);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_SUNKENOUTER, BF_RECT);

    //  
   {c := $C0C0C0;
    for y := r.Bottom - 2 downto r.Top + 1 do begin
      FBmp.Canvas.Pen.Color := c;
      FBmp.Canvas.MoveTo(xL + WS, y);
      FBmp.Canvas.LineTo(xR - WS - 1, y);
      c := c + $060606;
      if c > $FFFFFF then
        c := $FFFFFF;
    end;}
    if _prop_Gradient then
      GradientRect(r);

    //  
    x := xL + XSR;
    r := MakeRect(x, YSR, x + WR, h - YSR);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xR - XSR - WR - 1;
    r := MakeRect(x, YSR, x + WR, h - YSR);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);

    //  
    x := xC;
    r := MakeRect(x, YCR, x + WR, h - YCR);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xC - GCR - WR;
    r := MakeRect(x, YCR, x + WR, h - YCR);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xC + WR + GCR;
    r := MakeRect(x, YCR, x + WR, h - YCR);
    DrawEdge(FBmp.Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
  end;

  FBmp.Draw(DC,0,0);
end;

procedure THIZoomer.Resize;
begin
  FBmp.Width := Control.Width;
  FBmp.Height := Control.Height;
  ModelToSlider;
end;

procedure THIZoomer.onMouseDown(Sender: PControl; var Mouse: TMouseEventData);
begin
  FPressedType        := GetPressedType(Mouse.X);
  FPressedX           := Mouse.X;
  FPressedSliderX     := FSliderX;
  FPressedSliderWidth := FSliderWidth;
  FPressedPosition    := FModelPosition;
  FPressedRange       := FModelRange;
  FTimer.Enabled      := (FPressedType >= zptDown) and (FPressedType <= zptPageUp);
  FTimeStartRepeat    := GetTickCount + FRepeatDelay;
  Control.Invalidate;
end;

procedure THIZoomer.onMouseUp(Sender: PControl; var Mouse: TMouseEventData);
begin
  FTimer.Enabled := False;
  if FPressedType = GetPressedType(Mouse.X) then
    ChangePositionByPressedType;
  FPressedType := zptNone;
  Control.Cursor := crDefault;
  Control.Invalidate;
end;

procedure THIZoomer.onMouseMove(Sender: PControl; var Mouse: TMouseEventData);
var
  px, pw, mw, d: Integer;

  procedure ChangeX;
  var
    mp, mx: Integer;
  begin
    mp := FModelMaxRange - FPressedRange;
    mx := mw - FSliderWidth;
    if mx > 0 then
      FPressedPosition := MulDiv(FSliderX - Control.Height, mp, mx)
    else
      FPressedPosition := 0;
  end;

  procedure ChangeW;
  begin
    if mw > 0 then
      FPressedRange := MulDiv(FSliderWidth, FModelMaxRange, mw);
  end;

  procedure ChangeD;
  var
    p, r: Integer;
  begin
    p              := FModelPosition;
    r              := FModelRange;
    FModelPosition := FPressedPosition;
    FModelRange    := FPressedRange;
    VerifyData;
    if (p <> FModelPosition) or (r <> FModelRange) then
      DoChanged;
  end;

begin
  //if Mouse.Button <> mbLeft then Exit;

  mw := Control.Width - 2 * Control.Height;
  if FPressedType = zptNone then begin
    if GetPressedType(Mouse.X) > zptSlider then
      Control.Cursor := LoadCursor(0,IDC_SIZEWE)
    else
      Control.Cursor := crDefault;
  end
  else if mw >= FSliderMinWidth then begin
    px := FSliderX;
    pw := FSliderWidth;
    if FPressedType = zptSlider then begin
      FSliderX := FPressedSliderX - (FPressedX - Mouse.X);
      if (FSliderX + FSliderWidth) > (Control.Width - Control.Height) then
        FSliderX := Control.Width - Control.Height - FSliderWidth;
      if FSliderX < Control.Height then
        FSliderX := Control.Height;
      ChangeX;
      ChangeD;
    end
    else if FPressedType = zptSliderDown then begin
      FSliderX := FPressedSliderX - (FPressedX - Mouse.X);
      if FSliderX < Control.Height then
        FSliderX := Control.Height;
      d := FPressedSliderX + FPressedSliderWidth - FSliderMinWidth;
      if FSliderX > d then
        FSliderX := d;
      FSliderWidth := FPressedSliderX + FPressedSliderWidth - FSliderX;
      VerifySliderWidth;
      ChangeX;
      ChangeW;
      ChangeD;
    end
    else if FPressedType = zptSliderUp then begin
      FSliderWidth := FPressedSliderWidth + (Mouse.X - FPressedX);
      d := Control.Width - Control.Height - FSliderX;
      if FSliderWidth > d then
        FSliderWidth := d;
      VerifySliderWidth;
      ChangeW;
      ChangeD;
    end;
    if (px <> FSliderX) or (pw <> FSliderWidth) then
      Control.Invalidate;
  end;
end;

procedure THIZoomer.CreateBrushPattern(Bitmap: KOL.PBitmap);
var
  X, Y: Integer;
begin
  Bitmap.Width := 8;
  Bitmap.Height := 8;
  with Bitmap.Canvas{$ifndef F_P}^{$endif} do
  begin
    Brush.BrushStyle := bsSolid;
    Brush.Color := Color2RGB(_prop_Color);
    FillRect(MakeRect(0, 0, Control.Width, Control.Height));
    for Y := 0 to 7 do
      for X := 0 to 7 do
        if (Y mod 2) = (X mod 2) then  { toggles between even/odd pixles }
          Pixels[X, Y] := clWhite;     { on even/odd rows }
  end;
end;

procedure THIZoomer.Init;
var
  d: Cardinal;
begin
  Control := NewPaintBox(FParent);
  inherited;
  Control.TabStop        := False;
  Control.DoubleBuffered := True;
  Control.Height := GetSystemMetrics(SM_CYHSCROLL);

  FBmp := NewBitmap(Control.Width,Control.Height);
  FBmp.Canvas.Font.Assign(Control.Font);
  FBmp.Canvas.Font.FontName := 'WebDings';

  FModelPosition := _prop_Position;
  FModelMaxRange := _prop_MaxRange;
  FModelMinRange := _prop_MinRange;
  FModelRange    := _prop_Range;
  ModelToSlider;

  FGlyphs := NewBitmap(8,8);
  //FGlyphs.LoadFromResourceName(HInstance, 'ZOOMERGLYPHS');
  CreateBrushPattern(FGlyphs);

  FTimer         := NewTimer(200);
  FTimer.Enabled := False;
  FTimer.OnTimer := OnTimer;
  SystemParametersInfo(SPI_GETKEYBOARDDELAY, 0, @d, 0);
  if d > 3 then
    d := 3;
  FRepeatDelay := 250 * (d + 1);
  SystemParametersInfo(SPI_GETKEYBOARDSPEED, 0, @d, 0);
  if d > 31 then
    d := 31;
  FTimer.Interval := 400 - 12 * d;

  Control.OnPaint := Paint;
  Control.OnResize := Resize;
  Control.onMouseDown := onMouseDown;
  Control.onMouseUp := onMouseUp;
  Control.onMouseMove := onMouseMove;

  FShowSlider    := True;
  //_prop_Gradient := True;
end;

destructor THIZoomer.Destroy;
begin
  FGlyphs.Free;
  inherited Destroy;
end;

procedure THIZoomer.SetData(aMinRange, aMaxRange, aPosition, aRange: Integer);
begin
  FModelMinRange := aMinRange;
  FModelMaxRange := aMaxRange;
  FModelPosition := aPosition;
  FModelRange    := aRange;
  FShowSlider    := True;
  VerifyData;
  ModelToSlider;
  Control.Invalidate;
end;

procedure THIZoomer.HideSlider;
begin
  FShowSlider := False;
  Control.Invalidate;
end;

end.
